# type: ignore

from invoke import task, Collection


@task
def install(ctx, update=False):
    """Install the tools needed to build the docs."""
    if update:
        ctx.run('pip install -U -r doc/requirements.txt')
    elif not ctx.run('which sphinx-build', hide=True, warn=True):
        ctx.run('pip install -r doc/requirements.txt')


@task(install)
def clean(ctx):
    """Clean up the doc build directory."""
    ctx.run('make -C doc clean')


@task(install)
def build(ctx):
    """Build the HTML docs."""
    ctx.run('make -C doc html')


@task(install, build)
def open(ctx):
    """Open the docs in a browser (on macOS)."""
    ctx.run('open doc/build/html/index.html')


ns = Collection(install, clean, open)
ns.add_task(build, default=True)




from typing import Union, Tuple, Sequence

from .. import (ServerMechanism, ClientMechanism, ServerChallenge,
                ChallengeResponse)
from ..creds.client import ClientCredentials
from ..creds.plain import PlainCredentials
from ..exception import UnexpectedChallenge

__all__ = ['LoginMechanism']


class LoginMechanism(ServerMechanism, ClientMechanism):
    """Implements the LOGIN authentication mechanism."""

    def __init__(self, name: Union[str, bytes] = b'LOGIN') -> None:
        super().__init__(name)

    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[PlainCredentials, None]:
        try:
            first = responses[0]
        except (IndexError, ValueError) as exc:
            raise ServerChallenge(b'Username:') from exc
        try:
            second = responses[1]
        except (IndexError, ValueError) as exc:
            raise ServerChallenge(b'Password:') from exc
        username = first.response.decode('utf-8')
        password = second.response.decode('utf-8')
        return PlainCredentials(username, password, username), None

    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        if len(challenges) == 0:
            return ChallengeResponse(b'', b'')
        elif len(challenges) == 1:
            username = creds.authcid.encode('utf-8')
            return ChallengeResponse(challenges[0].data, username)
        elif len(challenges) == 2:
            password = creds.secret.encode('utf-8')
            return ChallengeResponse(challenges[1].data, password)
        raise UnexpectedChallenge()




from __future__ import absolute_import

import unittest

from pysasl import SASLAuth, ServerChallenge, ChallengeResponse
from pysasl.creds.client import ClientCredentials
from pysasl.creds.external import ExternalCredentials
from pysasl.exception import ExternalVerificationRequired, UnexpectedChallenge
from pysasl.identity import ClearIdentity
from pysasl.mechanisms.external import ExternalMechanism


class TestExternalMechanism(unittest.TestCase):

    def setUp(self) -> None:
        self.mech = ExternalMechanism()

    def test_availability(self) -> None:
        sasl = SASLAuth.defaults()
        self.assertIsNone(sasl.get_server(b'EXTERNAL'))
        self.assertIsNone(sasl.get_client(b'EXTERNAL'))
        sasl = SASLAuth.named([b'EXTERNAL'])
        self.assertEqual(self.mech, sasl.get_server(b'EXTERNAL'))
        self.assertEqual(self.mech, sasl.get_client(b'EXTERNAL'))
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)

    def test_server_attempt_issues_challenge(self) -> None:
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'', exc.data)
        else:
            self.fail('ServerChallenge not raised')

    def test_server_attempt_successful(self) -> None:
        result, final = self.mech.server_attempt([
            ChallengeResponse(b'', b'testuser')])
        self.assertIsNone(final)
        self.assertIsInstance(result, ExternalCredentials)
        self.assertEqual('testuser', result.authzid)
        self.assertEqual('', result.authcid)
        with self.assertRaises(ExternalVerificationRequired) as exc:
            result.verify(ClearIdentity('testuser', 'testpass'))
        self.assertIsNone(exc.exception.token)
        with self.assertRaises(ExternalVerificationRequired) as exc:
            result.verify(None)
        self.assertIsNone(exc.exception.token)

    def test_server_attempt_successful_empty(self) -> None:
        result, _ = self.mech.server_attempt([
            ChallengeResponse(b'', b'')])
        self.assertEqual('', result.authzid)
        self.assertEqual('', result.authcid)

    def test_client_attempt(self) -> None:
        creds = ClientCredentials('', '', 'testzid')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'testzid', resp1.response)
        resp2 = self.mech.client_attempt(creds, [ServerChallenge(b'')])
        self.assertEqual(b'testzid', resp2.response)
        self.assertRaises(UnexpectedChallenge,
                          self.mech.client_attempt,
                          creds, [ServerChallenge(b'')]*2)




from abc import abstractmethod
from typing import Optional, Sequence
from typing_extensions import Protocol

from . import Credentials
from ..identity import Identity

__all__ = ['ServerCredentials']


class ServerCredentials(Credentials, Protocol):
    """Credentials that are received from a client and should be authenticated
    against a known secret value.

    """

    __slots__: Sequence[str] = []

    @abstractmethod
    def verify(self, identity: Optional[Identity]) -> bool:
        """Authenticates the credentials against the given *identity*.

        Args:
            identity: The identity being authenticated.

        Raises:
            MechanismUnusable: The mechanism is not capable of verifying
                *identity*.

        """
        ...




import secrets
from typing import Optional

from .server import ServerCredentials
from ..identity import Identity
from ..prep import prepare

__all__ = ['PlainCredentials']


class PlainCredentials(ServerCredentials):
    """Implementation of :class:`~pysasl.creds.server.ServerCredentials` for
    typical SASL mechanisms like
    :class:`~pysasl.mechanisms.plain.PlainMechanism`.

    Args:
        authcid: Authentication ID string (the username).
        secret: Secret string (the password).
        authzid: Authorization ID string, if provided.

    """

    __slots__ = ['_authcid', '_secret', '_authzid']

    def __init__(self, authcid: str, secret: str, authzid: str = '') -> None:
        super().__init__()
        self._authcid = authcid
        self._secret = secret
        self._authzid = authzid or prepare(authcid)

    @property
    def authcid(self) -> str:
        return self._authcid

    @property
    def authzid(self) -> str:
        return self._authzid

    def verify(self, identity: Optional[Identity]) -> bool:
        if identity is not None:
            self_authcid = prepare(self.authcid)
            other_authcid = prepare(identity.authcid)
            return secrets.compare_digest(self_authcid, other_authcid)  \
                and identity.compare_secret(self._secret)
        return False

    def __repr__(self) -> str:
        return f'PlainCredentials({self.authcid!r}, ..., {self.authzid!r})'




from __future__ import absolute_import

import base64
import unittest

from pysasl.creds.plain import PlainCredentials
from pysasl.hashing import BuiltinHash, Cleartext
from pysasl.identity import ClearIdentity, HashedIdentity

builtin_hash = BuiltinHash(rounds=1000)

salt_sha256 = 'bzstsT0hfnnXDUPTJqbkhQ=='
password_sha256 = 'bzstsT0hfnnXDUPTJqbkhXVry+G06Vnz6GwNnrYvpAo1LBeLH8MIlufEh' \
    'NdKeFYd'
password_sha512 = 'EzkVJRnzPma/FYN2JM5XVj9oDl0qJwClAWywh8XAWz4iugQKMvlFPbyxM' \
    'HGWa9uIz16LC+aMMCYJT/Z78DR1wioV6eOdX8vgegxiKW8VWZk='


class TestCreds(unittest.TestCase):

    def test_hashed_builtin_good(self) -> None:
        creds = PlainCredentials('username', 'password')
        stored = HashedIdentity('username', password_sha256, hash=builtin_hash)
        self.assertEqual(password_sha256, stored.digest)
        self.assertTrue(creds.verify(stored))
        self.assertIsNone(stored.get_clear_secret())

    def test_hashed_builtin_invalid(self) -> None:
        creds = PlainCredentials('username', 'invalid')
        stored = HashedIdentity('username', password_sha256, hash=builtin_hash)
        self.assertFalse(creds.verify(stored))

    def test_hashed_builtin_copy(self) -> None:
        creds = PlainCredentials('username', 'password')
        builtin_copy = builtin_hash.copy()
        stored = HashedIdentity.create('username', 'password',
                                       hash=builtin_copy)
        self.assertTrue(creds.verify(stored))
        builtin_copy = builtin_hash.copy()
        stored = HashedIdentity('username', password_sha256, hash=builtin_copy)
        self.assertTrue(creds.verify(stored))
        builtin_copy = builtin_hash.copy(hash_name='sha512')
        stored = HashedIdentity('username', password_sha512, hash=builtin_copy)
        self.assertTrue(creds.verify(stored))

    def test_builtin_hash(self) -> None:
        salt = base64.b64decode(salt_sha256)
        self.assertEqual(password_sha256,
                         builtin_hash.hash('password', salt))

    def test_cleartext_good(self) -> None:
        creds = PlainCredentials('username', 'password')
        stored = ClearIdentity('username', 'password')
        self.assertTrue(creds.verify(stored))

    def test_cleartext_invalid(self) -> None:
        creds = PlainCredentials('username', 'password')
        self.assertFalse(creds.verify(ClearIdentity('username', 'invalid')))
        self.assertFalse(creds.verify(ClearIdentity('invalid', 'password')))

    def test_cleartext_copy(self) -> None:
        creds = PlainCredentials('username', 'password')
        stored = HashedIdentity('username', 'password',
                                hash=Cleartext())
        self.assertTrue(creds.verify(stored))
        stored = HashedIdentity('username', 'password',
                                hash=stored.hash.copy())
        self.assertTrue(creds.verify(stored))

    def test_cleartext_hash(self) -> None:
        self.assertEqual('password', Cleartext().hash('password'))

    def test_none(self) -> None:
        creds = PlainCredentials('username', 'password')
        self.assertFalse(creds.verify(None))

    def test_get_clear_secret(self) -> None:
        clear1 = ClearIdentity('username', 'password')
        self.assertEqual('password', clear1.get_clear_secret())
        clear2 = HashedIdentity.create('username', 'password',
                                       hash=Cleartext())
        self.assertEqual('password', clear2.get_clear_secret())
        non_clear = HashedIdentity.create('username', 'password',
                                          hash=builtin_hash.copy())
        self.assertIsNone(non_clear.get_clear_secret())



# type: ignore

from invoke import task, Collection

from .check import check_import


@task(check_import)
def mypy(ctx):
    """Run the mypy type checker."""
    ctx.run('mypy {} test'.format(ctx.package))


@task(check_import)
def pyright(ctx):
    """Run the pyright linter."""
    ctx.run('pyright {} test'.format(ctx.package))


@task(mypy, pyright)
def all(ctx):
    """Run all the type checker tools."""
    pass


ns = Collection(mypy, pyright)
ns.add_task(all, default=True)




from __future__ import absolute_import

import unittest
import email.utils

from unittest.mock import patch, Mock

from pysasl import SASLAuth, ServerChallenge, ChallengeResponse
from pysasl.creds.client import ClientCredentials
from pysasl.exception import (InvalidResponse, MechanismUnusable,
                              UnexpectedChallenge)
from pysasl.hashing import BuiltinHash
from pysasl.identity import ClearIdentity, HashedIdentity
from pysasl.mechanisms.crammd5 import CramMD5Result, CramMD5Mechanism

builtin_hash = BuiltinHash(rounds=1000)


class TestCramMD5Mechanism(unittest.TestCase):

    def setUp(self) -> None:
        self.mech = CramMD5Mechanism()

    def test_availability(self) -> None:
        sasl = SASLAuth.defaults()
        self.assertIsNone(sasl.get_server(b'CRAM-MD5'))
        self.assertIsNone(sasl.get_client(b'CRAM-MD5'))
        sasl = SASLAuth.named([b'CRAM-MD5'])
        self.assertEqual(self.mech, sasl.get_server(b'CRAM-MD5'))
        self.assertEqual(self.mech, sasl.get_client(b'CRAM-MD5'))
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)

    def test_result_verify_impossible(self) -> None:
        result = CramMD5Result('testuser', b'', b'')
        identity = HashedIdentity('testuser', 'digest',
                                  hash=builtin_hash.copy())
        with self.assertRaises(MechanismUnusable):
            result.verify(identity)

    @patch.object(email.utils, 'make_msgid')
    def test_server_attempt_issues_challenge(
            self, make_msgid_mock: Mock) -> None:
        make_msgid_mock.return_value = '<abc123.1234@testhost>'
        with self.assertRaises(ServerChallenge) as raised:
            self.mech.server_attempt([])
        self.assertEqual(b'<abc123.1234@testhost>', raised.exception.data)

    @patch.object(email.utils, 'make_msgid')
    def test_server_attempt_bad_response(self, make_msgid_mock: Mock) -> None:
        make_msgid_mock.return_value = '<abc123.1234@testhost>'
        self.assertRaises(InvalidResponse,
                          self.mech.server_attempt,
                          [ChallengeResponse(b'', b'testing')])

    @patch.object(email.utils, 'make_msgid')
    def test_server_attempt_successful(self, make_msgid_mock: Mock) -> None:
        make_msgid_mock.return_value = '<abc123.1234@testhost>'
        response = b'testuser 3a569c3950e95c490fd42f5d89e1ef67'
        result, final = self.mech.server_attempt([
            ChallengeResponse(b'<abc123.1234@testhost>', response)])
        self.assertIsNone(final)
        self.assertIsInstance(result, CramMD5Result)
        self.assertEqual('testuser', result.authcid)
        self.assertEqual('testuser', result.authzid)
        self.assertTrue(result.verify(ClearIdentity('testuser', 'testpass')))
        self.assertTrue(result.verify(ClearIdentity('testuser', 'testpass')))
        self.assertFalse(result.verify(ClearIdentity('testuser', 'badpass')))
        self.assertFalse(result.verify(ClearIdentity('baduser', 'testpass')))
        self.assertFalse(result.verify(None))

    def test_client_attempt(self) -> None:
        creds = ClientCredentials('testuser', 'testpass')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'', resp1.response)
        resp2 = self.mech.client_attempt(creds, [
            ServerChallenge(b'<abc123.1234@testhost>')])
        self.assertEqual(b'testuser 3a569c3950e95c490fd42f5d89e1ef67',
                         resp2.response)
        self.assertRaises(UnexpectedChallenge,
                          self.mech.client_attempt,
                          creds, [ServerChallenge(b'')]*2)




import re
from typing import Union, Tuple, Sequence

from .. import (ServerMechanism, ClientMechanism, ServerChallenge,
                ChallengeResponse)
from ..creds.client import ClientCredentials
from ..creds.external import ExternalCredentials
from ..exception import InvalidResponse, UnexpectedChallenge

__all__ = ['OAuth2Mechanism']


class OAuth2Mechanism(ServerMechanism, ClientMechanism):
    """Implements the `XOAUTH2`_ authentication mechanism, used by `OAuth 2.0`_
    systems to authenticate using access tokens.

    .. _XOAUTH2: https://developers.google.com/gmail/xoauth2_protocol
    .. _OAuth 2.0: https://tools.ietf.org/html/rfc6749

    """

    _pattern = re.compile(br'^user=(.*?)\x01auth=[bB][eE][aA][rR][eE][rR] '
                          br'(.*?)\x01\x01$')

    def __init__(self, name: Union[str, bytes] = b'XOAUTH2') -> None:
        super().__init__(name)

    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[ExternalCredentials, None]:
        try:
            first = responses[0]
        except IndexError as exc:
            raise ServerChallenge(b'') from exc

        match = re.match(self._pattern, first.response)
        if not match:
            raise InvalidResponse()
        user, token = match.groups()

        user_str = user.decode('utf-8')
        token_str = token.decode('utf-8')
        return ExternalCredentials(user_str, token_str), None

    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        if len(challenges) == 0:
            challenge = b''
        elif len(challenges) == 1:
            challenge = challenges[0].data
        else:
            raise UnexpectedChallenge()
        if challenge != b'':
            response = b''
        else:
            user = creds.authcid.encode('utf-8')
            token = creds.secret.encode('utf-8')
            response = b''.join((b'user=', user, b'\x01auth=Bearer ', token,
                                 b'\x01\x01'))
        return ChallengeResponse(challenge, response)




from typing import Optional, Sequence
from typing_extensions import Final

__all__ = ['AuthenticationError', 'UnexpectedChallenge', 'InvalidResponse',
           'ExternalVerificationRequired', 'MechanismUnusable']


class AuthenticationError(Exception):
    """Indicates that authentication failed due to a protocol error unrelated
    to any provided credentials.

    """

    __slots__: Sequence[str] = []


class UnexpectedChallenge(AuthenticationError):
    """During client-side authentication, the SASL mechanism received an
    authentication challenge from the server that it did not expect.

    """

    __slots__: Sequence[str] = []

    def __init__(self) -> None:
        super().__init__('Unexpected auth challenge')


class InvalidResponse(AuthenticationError):
    """During server-side authentication, the SASL mechanism received an
    authentication response from the client that was invalid.

    """

    __slots__: Sequence[str] = []

    def __init__(self) -> None:
        super().__init__('Invalid auth response')


class ExternalVerificationRequired(AuthenticationError):
    """The credentials are structurally valid but require external
    verification.

    If *token* is ``None``, the credentials provided no additional information
    for verification. Otherwise, *token* should be verified and authorized for
    the :attr:`~pysasl.creds.AuthenticationCredentials.identity` from the
    credentials.

    Args:
        token: A bearer token, if required for verification.

    """

    __slots__: Sequence[str] = ['token']

    def __init__(self, token: Optional[str] = None) -> None:
        super().__init__()
        self.token: Final = token


class MechanismUnusable(AuthenticationError):
    """The mechanism cannot be used to authenticate the given identity. Usually
    this is due to an unsupported hashing algorithm used in the server-side
    authentication database.

    Args:
        name: The mechanism name that is unusable.

    """

    __slots__: Sequence[str] = []

    def __init__(self, name: str) -> None:
        super().__init__(f'{name} cannot authenticate this identity')




from typing import Optional, Sequence

from .server import ServerCredentials
from ..exception import ExternalVerificationRequired
from ..identity import Identity

__all__ = ['ExternalCredentials']


class ExternalCredentials(ServerCredentials):
    """Credentials that require external verification, rather than by a
    traditional hashing algorithm.

    Args:
        authzid: Authorization ID string, if applicable.
        token: A bearer token, if required for verification.

    """

    __slots__: Sequence[str] = ['_authzid', '_token']

    def __init__(self, authzid: str, token: Optional[str] = None) -> None:
        super().__init__()
        self._authzid = authzid
        self._token = token

    @property
    def authcid(self) -> str:
        return ''

    @property
    def authzid(self) -> str:
        return self._authzid

    def verify(self, identity: Optional[Identity]) -> bool:
        raise ExternalVerificationRequired(self._token)

    def __repr__(self) -> str:
        return f'ExternalCredentials({self.authzid}, ...)'



"""Provides an abstraction and several implementations for the ability to hash
and verify secrets.

"""

import os
import base64
import hashlib
import secrets
from abc import abstractmethod
from typing import TypeVar, Any, Optional, Sequence, Dict
from typing_extensions import Protocol, Final

try:
    from passlib.context import CryptContext
    from passlib.apps import custom_app_context
except ImportError as _exc:  # pragma: no cover
    CryptContext = None
    custom_app_context = None
    _passlib_import_exc = _exc

__all__ = ['HashT', 'HashInterface', 'BuiltinHash', 'Cleartext', 'get_hash']

#: Type variable for a :class:`HashInterface`.
HashT = TypeVar('HashT', bound='HashInterface')


class HashInterface(Protocol):
    """Defines a basic interface for hash implementations. This is specifically
    designed to be compatible with :mod:`passlib` hashes.

    """

    __slots__: Sequence[str] = []

    @abstractmethod
    def copy(self: HashT, **kwargs: Any) -> HashT:
        """Return a copy of the hash implementation. The *kwargs* may be used
        by some hashes to modify settings on the hash.

        Args:
            self: The hash object being copied.
            kwargs: Updated settings for the returned hash.

        """
        ...

    @abstractmethod
    def hash(self, secret: str) -> str:
        """Hash the *value* and return the digest.

        Args:
            secret: The string to hash.

        """
        ...

    @abstractmethod
    def verify(self, secret: str, hash: str) -> bool:
        """Check the *secret* against the given *hash*.

        Args:
            secret: The string to check.
            hash: The hashed digest string.

        """
        ...


class BuiltinHash(HashInterface):
    """Implements :class:`HashInterface` using the :func:`hashlib.pbkdf2_hmac`
    function and random salt.

    Args:
        hash_name: The hash name.
        salt_len: The length of the random salt.
        rounds: The number of hash rounds.

    """

    __slots__: Sequence[str] = ['hash_name', 'salt_len', 'rounds']

    def __init__(self, *, hash_name: str = 'sha256', salt_len: int = 16,
                 rounds: int = 1000000) -> None:
        super().__init__()
        self.hash_name: Final = hash_name
        self.salt_len: Final = salt_len
        self.rounds: Final = rounds

    def _set_unless_none(self, kwargs: Dict[str, Any],
                         key: str, val: Any) -> None:
        if val is not None:
            kwargs[key] = val

    def copy(self, *, hash_name: Optional[str] = None,
             salt_len: Optional[int] = None,
             rounds: Optional[int] = None,
             **kwargs: Any) -> 'BuiltinHash':
        """Return a copy of the hash implementation, possibly with updated
        parameters.

        Args:
            hash_name: The updated hash name.
            salt_len: The updated length of the random salt.
            rounds: The updated number of hash rounds.
            kwargs: Additional keyword arguments are ignored.

        """
        copy_kwargs: Dict[str, Any] = {'hash_name': self.hash_name,
                                       'salt_len': self.salt_len,
                                       'rounds': self.rounds}
        new_kwargs: Dict[str, Any] = {}
        self._set_unless_none(new_kwargs, 'hash_name', hash_name)
        self._set_unless_none(new_kwargs, 'salt_len', salt_len)
        self._set_unless_none(new_kwargs, 'rounds', rounds)
        if new_kwargs:
            copy_kwargs.update(new_kwargs)
            return BuiltinHash(**copy_kwargs)
        else:
            return self

    def _hash(self, secret: str, salt: bytes) -> bytes:
        value_b = secret.encode('utf-8')
        digest = hashlib.pbkdf2_hmac(
            self.hash_name, value_b, salt, self.rounds)
        return salt + digest

    def hash(self, secret: str, salt: Optional[bytes] = None) -> str:
        """Hash the *secret* and return the digest.

        Args:
            secret: The string to hash.
            salt: A salt value to use instead of a random value.

        """
        if salt is None:  # pragma: no cover
            salt = os.urandom(self.salt_len)
        digest = self._hash(secret, salt)
        return self.encode_digest(digest)

    def verify(self, secret: str, hash: str) -> bool:
        digest_b = self.decode_digest(hash)
        salt = digest_b[0:self.salt_len]
        value_digest = self._hash(secret, salt)
        return secrets.compare_digest(value_digest, digest_b)

    @classmethod
    def encode_digest(cls, digest: bytes) -> str:
        """Encode the digest into a string. This uses :func:`~base64.b64encode`
        by default, but can be overridden by subclasses.

        Args:
            digest: The raw digest bytestring.

        """
        return base64.b64encode(digest).decode('ascii')

    @classmethod
    def decode_digest(cls, digest: str) -> bytes:
        """Decode the digest from a string. This uses :func:`~base64.b64decode`
        by default, but can be overridden by subclasses.

        Args:
            digest: The encoded digest string.

        """
        return base64.b64decode(digest)

    def __repr__(self) -> str:
        return 'BuiltinHash(hash_name=%r, salt_len=%r, rounds=%r)' % \
            (self.hash_name, self.salt_len, self.rounds)


class Cleartext(HashInterface):
    """Implements :class:`HashInterface` with no hashing performed."""

    __slots__: Sequence[str] = []

    def copy(self, **kwargs: Any) -> 'Cleartext':
        return self

    def hash(self, secret: str) -> str:
        return secret

    def verify(self, secret: str, hash: str) -> bool:
        return secrets.compare_digest(secret, hash)

    def __repr__(self) -> str:
        return 'Cleartext()'


def get_hash(*, no_passlib: bool = False,
             passlib_config: Optional[str] = None) \
        -> HashInterface:  # pragma: no cover
    """Provide a secure, default :class:`HashInterface` implementation.

    If :mod:`passlib` is not available, a custom hash is always used based on
    :func:`hashlib.pbkdf2_hmac`. The *passlib_config* parameter is ignored.

    If :mod:`passlib` is available, a :class:`~passlib.context.CryptContext` is
    loaded from the *passlib_config* parameter. If *passlib_config* is
    ``None``, then :attr:`passlib.apps.custom_app_context` is returned.

    Args:
        no_passlib: If true, do not use :mod:`passlib` even if available.
        passlib_config: A passlib config file.

    """
    context: HashInterface
    if no_passlib:
        return BuiltinHash()
    elif passlib_config is not None:
        if CryptContext is not None:
            context = CryptContext.from_path(passlib_config)
        else:
            raise _passlib_import_exc
    elif custom_app_context is not None:
        context = custom_app_context.copy()
    else:
        context = BuiltinHash()
    return context




import re
import hmac
import hashlib
import secrets
import email.utils
from typing import Union, Optional, Tuple, Sequence

from .. import (ServerMechanism, ClientMechanism, ServerChallenge,
                ChallengeResponse)
from ..creds.client import ClientCredentials
from ..creds.server import ServerCredentials
from ..exception import InvalidResponse, MechanismUnusable, UnexpectedChallenge
from ..identity import Identity
from ..prep import prepare

__all__ = ['CramMD5Result', 'CramMD5Mechanism']


class CramMD5Result(ServerCredentials):
    """Because this mechanism uses hash algorithms to compare secrets, the
    :meth:`~CramMD5Mechanism.server_attempt` method returns this sub-class
    which overrides the :meth:`.verify` method.

    """

    __slots__: Sequence[str] = ['_username', '_challenge', '_digest']

    def __init__(self, username: str, challenge: bytes,
                 digest: bytes) -> None:
        super().__init__()
        self._username = username
        self._challenge = challenge
        self._digest = digest

    @property
    def authcid(self) -> str:
        return self._username

    @property
    def authzid(self) -> str:
        return self._username

    def verify(self, identity: Optional[Identity]) -> bool:
        if identity is None:
            return False
        clear_secret = identity.get_clear_secret()
        if clear_secret is None:
            raise MechanismUnusable('CRAM-MD5')
        secret_b = clear_secret.encode('utf-8')
        expected_hmac = hmac.new(secret_b, self._challenge, hashlib.md5)
        expected_digest = expected_hmac.hexdigest().encode('ascii')
        self_authcid = prepare(self.authcid)
        other_authcid = prepare(identity.authcid)
        return secrets.compare_digest(self_authcid, other_authcid) \
            and hmac.compare_digest(expected_digest, self._digest)


class CramMD5Mechanism(ServerMechanism, ClientMechanism):
    """Implements the CRAM-MD5 authentication mechanism.

    Warning:
        Although secure during transport, offering this mechanism can be
        dangerous, as it can have implications about how the credentials are
        stored server-side.

    """

    _pattern = re.compile(br'^(.*) ([^ ]+)$')

    def __init__(self, name: Union[str, bytes] = b'CRAM-MD5') -> None:
        super().__init__(name)

    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[CramMD5Result, None]:
        try:
            first = responses[0]
        except IndexError as exc:
            challenge = email.utils.make_msgid().encode('utf-8')
            raise ServerChallenge(challenge) from exc

        match = re.match(self._pattern, first.response)
        if not match:
            raise InvalidResponse()
        username, digest = match.groups()

        username_str = username.decode('utf-8')
        return CramMD5Result(username_str, first.challenge, digest), None

    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        if len(challenges) < 1:
            return ChallengeResponse(b'', b'')
        elif len(challenges) > 1:
            raise UnexpectedChallenge()
        challenge = challenges[0].data
        authcid = prepare(creds.authcid).encode('utf-8')
        secret = prepare(creds.secret).encode('utf-8')
        digest = hmac.new(secret, challenge, hashlib.md5).hexdigest()
        response = b' '.join((authcid, digest.encode('ascii')))
        return ChallengeResponse(challenge, response)




from __future__ import absolute_import

import unittest

from pysasl import SASLAuth, ServerChallenge, ChallengeResponse
from pysasl.creds.client import ClientCredentials
from pysasl.creds.plain import PlainCredentials
from pysasl.exception import InvalidResponse, UnexpectedChallenge
from pysasl.identity import ClearIdentity
from pysasl.mechanisms.plain import PlainMechanism


class TestPlainMechanism(unittest.TestCase):

    def setUp(self) -> None:
        self.mech = PlainMechanism()

    def test_availability(self) -> None:
        sasl = SASLAuth.defaults()
        self.assertEqual(self.mech, sasl.get_server(b'PLAIN'))
        self.assertEqual(self.mech, sasl.get_client(b'PLAIN'))
        sasl = SASLAuth.named([b'PLAIN'])
        self.assertEqual(self.mech, sasl.get_server(b'PLAIN'))
        self.assertEqual(self.mech, sasl.get_client(b'PLAIN'))
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)

    def test_server_attempt_issues_challenge(self) -> None:
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'', exc.data)
        else:
            self.fail('ServerChallenge not raised')

    def test_server_attempt_bad_response(self) -> None:
        self.assertRaises(InvalidResponse,
                          self.mech.server_attempt,
                          [ChallengeResponse(b'', b'abcdefghi')])

    def test_server_attempt_successful(self) -> None:
        result, final = self.mech.server_attempt([
            ChallengeResponse(b'', b'abc\x00def\x00ghi')])
        self.assertIsNone(final)
        self.assertIsInstance(result, PlainCredentials)
        self.assertEqual('abc', result.authzid)
        self.assertEqual('def', result.authcid)
        self.assertTrue(result.verify(ClearIdentity('def', 'ghi')))
        self.assertFalse(result.verify(ClearIdentity('def', 'invalid')))
        self.assertFalse(result.verify(ClearIdentity('invalid', 'ghi')))

    def test_client_attempt(self) -> None:
        creds = ClientCredentials('testuser', 'testpass', 'testzid')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'testzid\x00testuser\x00testpass', resp1.response)
        resp2 = self.mech.client_attempt(creds, [ServerChallenge(b'')])
        self.assertEqual(b'testzid\x00testuser\x00testpass', resp2.response)
        self.assertRaises(UnexpectedChallenge,
                          self.mech.client_attempt,
                          creds, [ServerChallenge(b'')]*2)
